# Install 
* see instructions in howto_install.

# How to reproduce
The scripts for reproduction are in this directory. At the beginning of each script there 
are a couple of parameters marked as to be changed by the user. Please follow the instructions
below to set these parameters. The scripts are run with `python3 scriptname.py`.

## 4.1.1 Academic example (Table 1)
* The name of the script is `academic-ex.py`.
* The wavenumber is controlled by `omega` and should be set according to Table 1.
* The variable `n_layers` controls the number of layers, which are called J in the paper. 
  Please set this variable as given in the Table.
* The list `alpha_jumps` contains the row in the header of the Table. Setting  
  `alpha_jump = alpha_jumps[1]` sets the strength of the discontinuity to `0.25`
* The damping is controlled by `gamma_damping`. Set `gamma_damping = 1` for adding 
  damping to the preconditioner and `gamma_damping = 0` otherwise.
* Start the script with `python3 academic-ex.py`. The residuals in the GMRES iteration 
  are given as output.

## 4.1.2 SH-waves in frequency domain 

### Table 2
* The name of the script is `SH-wave-mantle.py`.
* The wavenumber and number of layers are controlled as described in 4.1.1
* The variable `eps_perturb` gives the strength of the sound speed perturbation. This 
  is only needed in the later examples. Please keep/set `eps_perturb = eps_perturbs[0]`.
* The right hand side is described by `source`. Pick `source[0]` for a random source 
  and `source[1]` for a Dirac source. In case of a random source we have to set 
  `zero_source_layer0 = True` to set the right hand side to zero on the first layer.
* The boolean `use_PML` determines whether a PML is set at the Earth's surface.
  Set `use_PML = True` for a PML and `use_PML = False` for the free surface 
  boundary condition.
* Please keep/set `use_moving_PML = True` to use the moving PML approximation 
  of the DtN.

### Fig. 2
* To create vtk output one can set `create_vtk_output = True`. This will 
  create a file called `vtk_SH-wave` that contains the real part of the 
  solution computed with a direct solver.
  You can use different software to visualize the data from that vtk file,
  e.g. Paraview (Paraview has been used for the generation of Fig. 2).

### Fig. 3
* The residuals correspond to the column "Dirac source" in Table 2
  and can be reproduced as described above.

## 4.2 Tensor product DtN

### Table 3
* The same script `SH-wave-mantle.py` as in 4.1.2 is used.
* Set `use_moving_PML = False` to use the tensor product DtN.
* The right hand side is chosen as described in 4.1.2. In case of a random source please set 
  `zero_source_layer0 = False`.
* Set `use_PML = False`, because this test is for the free surface boundary condition.
* Please set `eps_perturb = eps_perturbs[0]`.
* The wavenumber and number of layers are controlled as described in 4.1.1
* The relative L2-error after one application is printed out to the terminal.

### 4.2.1 Sound speed perturbations (Table 4)
* The same script `SH-wave-mantle.py` as in 4.1.2 is used.
* The list `eps_perturbs` contains the row in the header of the Table. Set   
  `eps_perturb = eps_perturbs[1]` to perturb the sound speed by 0.0625% and so on.
* Set `source = sources[0]`and `zero_source_layer0 = False`. 
* Set `use_PML = False` to reproduce the first part of the Table with the free surface 
  boundary condition and `use_PML = True` for the second Table with the PML at the 
  Earth's surface. 
* Please set `use_moving_PML = True` to use the moving PML approximation 
  of the DtN or `use_moving_PML = False` for the tensor product DtN. 

## 5.1 Modal analysis (Fig. 4)
* The script is called `SH-wave-DtN-err.py`.
* Set `use_PML = True` for a PML and `use_PML = False` for the free surface 
  boundary condition.
* Execute the script with `python3 SH-wave-DtN-err.py`. 
* The relative error for each mode is printed for `omega = 512`. Press 
  any key to continue to `omega = 1024` and so on.

# General structure of the code 
The experiments are implemented via the `Python` interface of the finite element package `Netgen/NGSolve`. 
An extensive documentation is available [here](https://ngsolve.org/docu/v6.2.1910/). 
As described above, the three main files to reproduce the results are `academic-ex.py`,`SH-wave-mantle.py` 
and `SH-wave-DtN-err.py`. Some common functions have been outsourced to auxiliary files:

* `sweeping_geom.py` contains the function `Layered_Mesh` which creates a two-dimensional layered mesh as used in DOSM. Additionally, 
two functions for creating one-dimensional meshes, which are needed for solving the ODEs, are provided.
* `sweeping_helper.py` provides a function to load the coefficients of the PREM model and some other convenience functions.

Since the files `academic-ex.py` and `SH-wave-mantle.py` are very similar only the latter will be described here
in detail. The main steps have already been marked as comments in the code and will now be briefly described.

* Parameters: Here the main parameters for the experiment like the wavenumber, the number of layers and so on are defined. Please
adjust these parameters as specified above.
* Further parameters: Some further parameters like the order of the finite element spaces are defined. Note that modifying 
these parameters may lead to different results than in the paper.
* Layered mesh: The mesh for the 2D discretization is created by calling the imported function `Layered_Mesh` with the 
chosen parameters.
* Coefficients of problem: load the coefficients of the Earth model by calling the import function `load_PREM`.
* FEM space: Create the finite element space as well as Proxy and-GridFunctions. Please see the NGSolve documentation for details.
* Domain decomposition: Here the degrees of freedom (dofs) of the FEM space are partioned according to the layered decomposition 
of the mesh. They are stored into lists with descriptive names. For example `IF2dof` contains the dofs that are associated with 
the interfaces (IF) between two neighbouring subdomains.
* PML: If a PML at the Earth's surface should be used then we define it in this step. 
* Solution with direct solver: we compute this to check that the iterative solver actually converges to the correct solution.
* Computation of DtN numbers: Here we solve the ODEs as described in Proposition 1 of the paper in order to obtain the diagonal elements
of the DtN in the eigenbasis.
* Computation of layer matrices and factorizations: In the domain decomposition method we need to solve the linear systems 
on the subdomains. To this end, the matrices on the sobdomains are assembled and factorized in this step.
* Transmission operators of DOSM: Transmission operators (in our case the DtN) are responsible for propagating information between the subdomain problems. 
In this step these operators are declared. If `use_moving_PML = True` the function `DtN_moving_PML` will be used to approximate the DtN
by a moving PML. In the other case, the tensor product DtN implemented in the function `DtN_semi_discrete` is employed.
* Assembling the matrices for solving the system on layer interfaces: To this end, we apply the transmission operator to each dof on the 
respective interface and compute a factorization of the obtained matrix. 
* DOSM algorithm in matrix form: This is the implementation of Algorithm 2 in the paper "A CLASS OF ITERATIVE SOLVERS FOR THE HELMHOLTZ EQUATION:
FACTORIZATIONS. SWEEPING PRECONDITIONERS, SOURCE TRANSFER,SINGLE LAYER POTENTIALS, POLARIZED TRACES, AND OPTIMIZED SCHWARZ METHODS" (M.J. Gander and H. Zhang) 
* Class definition of preconditioner: define the preconditioner in the framework of `NGSolve`.
* One application of preconditioner: apply the preconditioner once to the right hand side and print the relative error. 
* Preconditioned GMRES iteration: Solve the problem using preconditioned GMRES and print the value of the residual in every step. 


