
from netgen.meshing import *
from netgen.csg import *
import ngsolve
from math import pi
import copy

def Layered_Mesh(r_layer,nx,quads,refine_layer,theta_bnd,x_max=pi):


    vertices_layer_nr = {}
    vertices_layer_coord = {}
    mesh_nodes = {}

    mesh = Mesh()
    mesh.dim=2

    mesh.Add(FaceDescriptor(surfnr=1,domin=1,bc=1))

    n_layers = len(r_layer) - 1

    slavej = []
    masterj = []

    pids_inner = []
    coords_inner = []
    for j in range(nx+1):
        x_inner,y_inner = x_max*j/nx, r_layer[0]
        coords_inner.append( (x_inner,y_inner))
        pids_inner.append(mesh.Add (MeshPoint(Pnt(x_inner,y_inner,0))))
        if j == 0:
            slavej.append(pids_inner[-1])
        if j == nx:
            masterj.append(pids_inner[-1])
        
    vertices_layer_nr[n_layers] = pids_inner
    vertices_layer_coord[n_layers] = coords_inner


    def MakeLayer(quads, nx,mesh,pids_next_lvl,r_start,r_end,mapping = None,refine_layer=False,domin=1,bc=1):

        fd_layer = mesh.Add(FaceDescriptor(surfnr=1,domin=domin,bc=1))

        pids_start = pids_next_lvl
        pids_middle = []
        pids_end = []
        if refine_layer:
            pids_refined_center = []
            pids_refined_end = []
        r_middle = r_start + (r_end-r_start)/2
        for j in range(nx+1):
            x_middle,y_middle = x_max*j/nx, r_middle
            x_end,y_end = x_max*j/nx, r_end
            if mapping:
                x_middle,y_middle = mapping(x_middle,y_middle)
                x_end,y_end = mapping(x_end,y_end)  
                
            node_nr_middle = mesh.Add (MeshPoint(Pnt(x_middle,y_middle,0)))
            pids_middle.append(node_nr_middle)    
            node_nr_end = mesh.Add (MeshPoint(Pnt(x_end,y_end,0)))
            pids_end.append(node_nr_end)  

            if j == 0:
                slavej.append(pids_middle[-1])
                slavej.append(pids_end[-1])
            if j == nx:
                masterj.append(pids_middle[-1])
                masterj.append(pids_end[-1])

            mesh_nodes[node_nr_middle.nr] = (x_middle,y_middle)
            mesh_nodes[node_nr_end.nr] = (x_end,y_end)

        if quads:
            for i in range(nx):
                elpids1 = [pids_start[i],pids_start[i+1],pids_middle[i+1],pids_middle[i] ]
                elpids2 = [pids_middle[i],pids_middle[i+1],pids_end[i+1],pids_end[i] ]
                mesh.Add(Element2D(fd_layer,elpids1)) 
                mesh.Add(Element2D(fd_layer,elpids2)) 
            for i in range(nx):
                mesh.Add(Element1D([pids_start[i], pids_start[i+1]], index=bc))
            mesh.SetBCName(bc-1,"layer"+str(domin)+"-bnd")
            return nx,pids_end

        if refine_layer:
            for j in range(nx):
                x_end_refined,y_end = x_max*j/nx + x_max*1/(2*nx),r_end
                x_center_refined,y_center_refined = x_max*j/nx + x_max*1/(2*nx) , r_middle + 0.5*(r_end-r_middle)
                if mapping:
                    x_end_refined,y_end = mapping(x_end_refined,y_end)
                    x_center_refined,y_center_refined = mapping(x_center_refined,y_center_refined)
                
                node_nr_end_refined = mesh.Add (MeshPoint(Pnt(x_end_refined,y_end,0)))
                pids_refined_end.append(node_nr_end_refined) 

                node_nr_center_refined = mesh.Add (MeshPoint(Pnt(x_center_refined,y_center_refined,0)))
                pids_refined_center.append(node_nr_center_refined) 

                mesh_nodes[node_nr_end_refined.nr] = (x_end_refined,y_end)
                mesh_nodes[node_nr_center_refined.nr]  = (x_center_refined,y_center_refined)
                                    
        
        for i in range(nx):


            elpids1 = [pids_start[i],pids_start[i+1],pids_middle[i] ]
            elpids2 = [pids_start[i+1],pids_middle[i+1],pids_middle[i] ]

            mesh.Add(Element2D(fd_layer,elpids1)) 
            mesh.Add(Element2D(fd_layer,elpids2))   

            if not refine_layer:
                elpids3 = [pids_middle[i],pids_middle[i+1],pids_end[i] ]
                elpids4 = [pids_middle[i+1],pids_end[i+1],pids_end[i] ]
                mesh.Add(Element2D(fd_layer,elpids3)) 
                mesh.Add(Element2D(fd_layer,elpids4))          
            else:
                elpids3 = [pids_middle[i],pids_middle[i+1],pids_refined_center[i] ]              
                elpids5 = [pids_middle[i+1],pids_end[i+1],pids_refined_center[i]]
                elpids7 = [pids_refined_center[i],pids_end[i+1],pids_refined_end[i]]

                elpids4 = [pids_middle[i],pids_refined_center[i],pids_end[i]]
                elpids6 = [pids_refined_center[i],pids_refined_end[i],pids_end[i]]
                
                mesh.Add(Element2D(fd_layer,elpids3)) 
                mesh.Add(Element2D(fd_layer,elpids4)) 
                mesh.Add(Element2D(fd_layer,elpids5)) 
                mesh.Add(Element2D(fd_layer,elpids6)) 
                mesh.Add(Element2D(fd_layer,elpids7)) 


        for i in range(nx):
            mesh.Add(Element1D([pids_start[i], pids_start[i+1]], index=bc))
            
        mesh.SetBCName(bc-1,"layer"+str(domin)+"-bnd")

        if refine_layer:
            pids_next_lvl = []
            for i in range(len(pids_end)):
                pids_next_lvl.append(pids_end[i])
                if i < len(pids_refined_end):
                    pids_next_lvl.append(pids_refined_end[i])
            return nx*2,pids_next_lvl
        else:
            return nx,pids_end


    pids_next_lvl = pids_inner

    for i in range(n_layers):

        #refine_layer = True if (i >= 4 and i <= 7) else False # Mesh-RR
        #refine_layer = True if (i >= 5 and i <= 8) else False # Mesh-R
        #refine_layer = False
        nx,pids_next_lvl = MakeLayer(quads=quads,nx=nx,mesh=mesh,pids_next_lvl=pids_next_lvl,r_start=r_layer[i],r_end=r_layer[i+1], 
                       mapping = None,
                       refine_layer=refine_layer[i],
                       domin = n_layers-i,
                       bc =  n_layers+1-i
              
                      )
    for i in range(nx):
        mesh.Add(Element1D([pids_next_lvl[i], pids_next_lvl[i+1]], index=1))

    if theta_bnd == 'periodic': 
        for j in range(len(slavej)):        
            mesh.AddPointIdentification(masterj[j],slavej[j],identnr=2,type=2) 
    else:
        for j in range(len(slavej)-1):  
            mesh.Add(Element1D([slavej[j], slavej[j+1]], index=n_layers+2))
            mesh.Add(Element1D([masterj[j], masterj[j+1]], index=n_layers+3))
            
        mesh.SetBCName(n_layers+1,"left")
        mesh.SetBCName(n_layers+2,"right")


    mesh.Compress()       
    mesh = ngsolve.Mesh(mesh)
    return mesh



from netgen.meshing import  Mesh as NetMesh
from ngsolve import Mesh as NGSMesh

def Make1DMesh(layer_nr,r_layer):
    
    mesh_1D = NetMesh(dim=1)
    pids = []   

    for nr,r_cord in enumerate(r_layer[-1-layer_nr:]):
        pids.append (mesh_1D.Add (MeshPoint(Pnt(r_cord, 0, 0))))  
        if nr + 1 < len(r_layer[-1-layer_nr:]):
            r_mid = r_cord + 0.5*(r_layer[-1-layer_nr:][nr+1] - r_cord )
            pids.append (mesh_1D.Add (MeshPoint(Pnt(r_mid, 0, 0))))  

    n_mesh = len(pids)-1
    for i in range(n_mesh):
        mesh_1D.Add(Element1D([pids[i],pids[i+1]],index=1))
    mesh_1D.Add (Element0D( pids[0], index=1))
    mesh_1D.Add (Element0D( pids[n_mesh], index=2))
    mesh_1D.SetBCName(0,"left")
    mesh_1D.SetBCName(1,"right")
    mesh_1D = NGSMesh(mesh_1D)    
    return mesh_1D

def Make1DMesh_refined(layer_nr,r_layer,nelem=100):
    
    mesh_1D = NetMesh(dim=1)
    pids = []   

    r_nodes = [ r_layer[-1-layer_nr] + j*(r_layer[-1] - r_layer[-1-layer_nr])/nelem for j in range(nelem +1)]
    
    #print("r_nodes =", r_nodes)

    for r_cord in r_nodes:
        pids.append (mesh_1D.Add (MeshPoint(Pnt(r_cord, 0, 0))))  

    n_mesh = len(pids)-1
    for i in range(n_mesh):
        mesh_1D.Add(Element1D([pids[i],pids[i+1]],index=1))
    mesh_1D.Add (Element0D( pids[0], index=1))
    mesh_1D.Add (Element0D( pids[n_mesh], index=2))
    mesh_1D.SetBCName(0,"left")
    mesh_1D.SetBCName(1,"right")
    mesh_1D = NGSMesh(mesh_1D)    
    return mesh_1D
