from netgen.meshing import *
from netgen.csg import *
from ngsolve import *
from math import pi
import matplotlib.colors as colors
import scipy.sparse.linalg
import numpy as np


def load_PREM():

    c = CoefficientFunction(0.0)
    rho = CoefficientFunction(0.0) 

    c_1D = CoefficientFunction(0.0)
    rho_1D = CoefficientFunction(0.0) 

    # load PREM coefficients 
     
    r_d,d_d,rho_d,_,_,_,c_d,_,_,_ = np.loadtxt("PREM.txt", unpack=True) 

    R_earth = 6371
    r_d = r_d / R_earth
    #r_d = r_d
    spline_order = 4

    r_d = r_d.tolist()[::-1]
    c_d = c_d.tolist()[::-1]
    rho_d = rho_d.tolist()[::-1]
    #print("r_d = ", r_d)

    for (ra,rb) in [(40,86),(87,90),(91,98),(99,119),(120,138),(139,171),(172,198)]:

        r_beg = r_d[ra]
        #print("r_beg = ", r_beg)
        c_beg = c_d[ra]
        #print("c_beg = ", c_beg)
        rho_beg = rho_d[ra]

        r_end = r_d[rb]
        #print("r_end = ", r_end)
        c_end = c_d[rb]
        #print("c_end = ", c_end)
        rho_end = rho_d[rb]
        #print(" ")

        r_d_slice = np.append([r_beg for i in range(spline_order)],r_d[ra:rb+1] )
        c_d_slice = np.append([c_beg for i in range(spline_order)],c_d[ra:rb+1] )
        rho_d_slice = np.append([rho_beg for i in range(spline_order)],rho_d[ra:rb+1] )

        r_d_slice = np.append(r_d_slice,[r_end for i in range(spline_order)])
        c_d_slice = np.append(c_d_slice,[c_end for i in range(spline_order)])
        rho_d_slice = np.append(rho_d_slice,[rho_end for i in range(spline_order)])

        r_d_slice = r_d_slice.tolist()
        c_d_slice = c_d_slice.tolist()
        rho_d_slice = rho_d_slice.tolist()
        
        c_in = BSpline(spline_order,r_d_slice,c_d_slice)
        rho_in = BSpline(spline_order,r_d_slice,rho_d_slice)

        c += IfPos(y-r_beg,IfPos(r_end-y,c_in(y),0),0)
        c_1D += IfPos(x-r_beg,IfPos(r_end-x,c_in(x),0),0)
        
        rho += IfPos(y-r_beg,IfPos(r_end-y,rho_in(y),0),0)
        rho_1D += IfPos(x-r_beg,IfPos(r_end-x,rho_in(x),0),0)

    return c,c_1D,rho,rho_1D

def BitArrayFromList(l,mesh,fes, elements = True):
    if elements:
        ba = BitArray(mesh.ne)
    else:
        ba = BitArray(fes.ndof)
    ba.Clear()
    for el in l:
        ba[el] = True
    return ba

def P_DoFs(M,test,basis):
    return (M[test,:][:,basis])
